from typing import Union

from pyattest.configs.config import Config
from pyattest.exceptions import PyAttestException


class Attestation:
    def __init__(self, raw: Union[bytes, str], nonce: bytes, config: Config):
        self.raw = raw  # The raw attestation the verifier expects
        self.nonce = nonce
        self.config = config

        self.verified = False
        self._verified_data = {}

    @property
    def data(self):
        if not self.verified:
            raise PyAttestException('Attestation needs to be verified before accessing it.')

        return self._verified_data

    def verify(self):
        self.config.verifier_class(self).verify()

    def verified_data(self, data: dict):
        """ The verifier from the config can set all relevant data once the verification is complete. """
        self.verified = True
        self._verified_data = data



from abc import ABC, abstractmethod

from pyattest.attestation import Attestation


class Verifier(ABC):
    def __init__(self, attestation: Attestation):
        self.attestation = attestation

    @abstractmethod
    def verify(self):
        ...



class PyAttestException(Exception):
    pass


class InvalidJWT(PyAttestException):
    pass


class InvalidBasicIntegrity(PyAttestException):
    pass


class InvalidCtsProfile(PyAttestException):
    pass


class ExtensionNotFoundException(PyAttestException):
    pass


class InvalidNonceException(PyAttestException):
    pass


class InvalidKeyIdException(PyAttestException):
    pass


class InvalidAppIdException(PyAttestException):
    pass


class InvalidCounterException(PyAttestException):
    pass


class InvalidAaguidException(PyAttestException):
    pass


class InvalidCredentialIdException(PyAttestException):
    pass


class InvalidCertificateChainException(PyAttestException):
    pass



import datetime

from cryptography import x509
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.x509.oid import NameOID

key_usage = x509.KeyUsage(
    True,  # digital_signature
    False,  # content_commitment
    False,  # key_encipherment
    False,  # data_encipherment
    False,  # key_agreement
    True,  # key_cert_sign
    True,  # crl_sign
    False,  # encipher_only
    False  # decipher_only
)


def generate():
    """ Generate root and intermediate certificates for the apple attestation. """
    root_private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048, )

    with open('pyattest/testutils/fixtures/root_key.pem', 'wb') as f:
        f.write(root_private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.BestAvailableEncryption(b'123'),
        ))

    subject = issuer = x509.Name([x509.NameAttribute(NameOID.ORGANIZATION_NAME, 'pyattest-testing-ca')])
    cert = x509.CertificateBuilder() \
        .subject_name(subject) \
        .issuer_name(issuer) \
        .public_key(root_private_key.public_key()) \
        .serial_number(x509.random_serial_number()) \
        .not_valid_before(datetime.datetime.utcnow()) \
        .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=10)) \
        .add_extension(key_usage, critical=False) \
        .sign(root_private_key, hashes.SHA256())

    with open('pyatest/testutils/fixtures/root_cert.pem', 'wb') as f:
        f.write(cert.public_bytes(serialization.Encoding.PEM))



from pathlib import Path

from pyattest.configs.config import Config
from pyattest.verifiers.apple import AppleVerifier


class AppleConfig(Config):
    verifier_class = AppleVerifier

    def __init__(self, key_id: bytes, app_id: str, production: bool = False, root_ca: bytes = None):
        self.key_id = key_id
        self.app_id = app_id
        self.production = production
        self._custom_root_ca = root_ca

    @property
    def oid(self) -> str:
        """ Object Identifier of the certificate extension containing our nonce. """
        return '1.2.840.113635.100.8.2'

    @property
    def root_ca(self) -> bytes:
        """
        Apples App Attestation Root CA. This can be overwritten for easier testing.

        See also: https://www.apple.com/certificateauthority/private/
        """
        if self._custom_root_ca:
            return self._custom_root_ca

        folder = Path(__file__).parent / '../certificates'
        return Path(folder / 'Apple_App_Attestation_Root_CA.pem').read_bytes()



import base64
import datetime
from pathlib import Path

import jwt
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization.base import load_pem_private_key
from cryptography.x509.base import load_pem_x509_certificate
from cryptography.x509.oid import NameOID

from pyattest.testutils.factories.certificates import key_usage


def get(apk_package_name: str, nonce: bytes, basic_integrity: bool = True, cts_profile: bool = True,
        apk_cert_digest: bytes = None):
    """ Helper to create a fake google attestation. """
    root_key = load_pem_private_key(Path('pyattest/testutils/fixtures/root_key.pem').read_bytes(), b'123')
    root_cert = load_pem_x509_certificate(Path('pyattest/testutils/fixtures/root_cert.pem').read_bytes())
    apk_cert_digest = apk_cert_digest or b'foobar'

    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
    )

    subject = x509.Name([x509.NameAttribute(NameOID.ORGANIZATION_NAME, 'pyattest-testing-leaf')])
    cert = x509.CertificateBuilder() \
        .subject_name(subject) \
        .issuer_name(root_cert.subject) \
        .public_key(private_key.public_key()) \
        .serial_number(x509.random_serial_number()) \
        .not_valid_before(datetime.datetime.utcnow()) \
        .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=10)) \
        .add_extension(key_usage, critical=False) \
        .add_extension(x509.SubjectAlternativeName([x509.DNSName('pyattest-testing-leaf.ch')]), critical=False) \
        .sign(root_key, hashes.SHA256())

    data = {
        'timestampMs': 9860437986543,
        'nonce': base64.b64encode(nonce).decode(),
        'apkPackageName': apk_package_name,
        'apkCertificateDigestSha256': [base64.b64encode(apk_cert_digest).decode()],
        'ctsProfileMatch': cts_profile,
        'basicIntegrity': basic_integrity,
        'evaluationType': 'BASIC'
    }

    headers = {'x5c': [
        cert.public_bytes(serialization.Encoding.PEM).decode().replace('-----BEGIN CERTIFICATE-----\n', '').replace(
            '\n-----END CERTIFICATE-----\n', ''),
        root_cert.public_bytes(serialization.Encoding.PEM).decode().replace('-----BEGIN CERTIFICATE-----\n',
                                                                            '').replace('\n-----END CERTIFICATE-----\n',
                                                                                        ''),
    ]}

    return jwt.encode(data, private_key, algorithm='RS256', headers=headers), apk_cert_digest



import struct
from hashlib import sha256

from asn1crypto.x509 import Certificate, Extension
from certvalidator import CertificateValidator, ValidationContext
from certvalidator.errors import PathValidationError, PathBuildingError
from certvalidator.path import ValidationPath

from pyattest.exceptions import ExtensionNotFoundException, InvalidNonceException, InvalidKeyIdException, \
    InvalidAppIdException, InvalidCounterException, InvalidAaguidException, InvalidCredentialIdException, \
    InvalidCertificateChainException
from pyattest.verifiers.verifier import Verifier
from cbor2 import loads as cbor_decode


class AppleVerifier(Verifier):
    def verify(self):
        """
        Verify the given attestation based on the Apple documentation. The attestation is CBOR encoded and after
        decoding contains all relevant data according to the Webauthn specification.

        Raises a PyAttestException as soon as one of the verification steps fails.

        See also:
        Apple Documentation https://developer.apple.com/documentation/devicecheck/validating_apps_that_connect_to_your_server
        Webauthn Specification https://www.w3.org/TR/webauthn/#fig-attStructs

        Relevant Data:
        rp_id              AppId SHA-256 digest
        counter            The number of times the app used the attested key to sign an assertion.
        aaguid             Environment
        credential_data    Variable length, contains the credential_id and credential_public_key. The length is
                           set at byte 17 and 18.
        """
        data = self.unpack(self.attestation.raw)

        chain = self.verify_certificate_chain(data['raw']['attStmt']['x5c'])
        self.verify_nonce(data['raw']['authData'], self.attestation.nonce, chain[-1])
        self.verify_key_id(chain[-1])
        self.verify_app_id(data['rp_id'])
        self.verify_counter(data['counter'])
        self.verify_aaguid(data['aaguid'])
        self.verify_credential_id(data['credential_id'], chain[-1])

        self.attestation.verified_data({'data': data, 'certs': chain})

    @staticmethod
    def unpack(raw: bytes) -> dict:
        """ Extract in `verify` method mentioned relevant data from cbor encoded raw bytes input. """
        raw = cbor_decode(raw)

        credential_data = raw['authData'][37:]
        credential_id_length = struct.unpack('!H', credential_data[16:18])[0]

        return {
            'raw': raw,
            'rp_id': raw['authData'][:32],
            'counter': struct.unpack('!I', raw['authData'][33:37])[0],
            'aaguid': credential_data[:16],
            'credential_id': credential_data[18:18 + credential_id_length],
        }

    @staticmethod
    def verify_credential_id(credential_id: bytes, cert: Certificate):
        """ Verify that the authenticator data’s credentialId field is the same as the key identifier. """
        if credential_id != cert.public_key.sha256:
            raise InvalidCredentialIdException

    def verify_aaguid(self, aaguid: bytes):
        """
        Verify that the authenticator data’s aaguid field is either appattestdevelop if operating in the development
        environment, or appattest followed by seven 0x00 bytes if operating in the production environment.

        See also: https://developer.apple.com/documentation/bundleresources/entitlements/com_apple_developer_devicecheck_appattest-environment
        """
        if self.attestation.config.production and aaguid == b'appattest\x00\x00\x00\x00\x00\x00\x00':
            return

        if not self.attestation.config.production and aaguid == b'appattestdevelop':
            return

        raise InvalidAaguidException

    @staticmethod
    def verify_counter(counter: int):
        """ Verify that the authenticator data’s counter field equals 0. """
        if counter != 0:
            raise InvalidCounterException

    def verify_app_id(self, rp_id: bytes):
        """
        Compute the SHA-256 hash of your app’s App ID, and verify that this is the same as the authenticator
        data’s RP ID hash.
        """
        if rp_id != sha256(self.attestation.config.app_id.encode()).digest():
            raise InvalidAppIdException

    def verify_key_id(self, cert: Certificate):
        """
        Create the SHA-256 hash of the public key in credCert, and verify that it matches the key
        identifier from your app.
        """
        expected_key_id = cert.public_key.sha256
        if expected_key_id != self.attestation.config.key_id:
            raise InvalidKeyIdException

    def verify_nonce(self, auth_data: bytes, nonce: bytes, cert: Certificate):
        """
        Create clientDataHash as the SHA-256 hash of the one-time challenge sent to your app before performing the
        attest_apple, and append that hash to the end of the authenticator data (authData from the decoded object).

        Generate a new SHA-256 hash of the composite item to create nonce.

        Obtain the value of the credCert extension with OID 1.2.840.113635.100.8.2, which is a DER-encoded ASN.1
        sequence. Decode the sequence and extract the single octet string that it contains. Verify that the string
        equals nonce.

        Interactive ASN.1 decoder: https://lapo.it/asn1js/

        Sequence:
            SEQUENCE (2 elem)
              OBJECT IDENTIFIER
              OCTET STRING (38 byte)
                SEQUENCE (1 elem)
                    OCTET STRING (32 byte)
        """
        client_data_hash = auth_data + sha256(nonce).digest()
        calculated_nonce = sha256(client_data_hash).digest()

        # TODO: It would be even nice if we could register the apple oid like they do it
        # with the pdf extension: https://github.com/wbond/asn1crypto/blob/79fa04ec6534c8aa000bf28e1f5bae7f2929bd1a/asn1crypto/pdf.py
        extension = self._get_extension(self.attestation.config.oid, cert)

        # asn1crypto somehow doesn't understand the sequence bellow the octet string, so we can just remove
        # the first 6 bytes to get the correct `expected_nonce`
        expected_nonce = extension[2].contents[6:]
        if calculated_nonce != expected_nonce:
            raise InvalidNonceException

    def verify_certificate_chain(self, chain: list) -> ValidationPath:
        """
        Verify that the x5c array contains the intermediate and leaf certificates for App Attest, starting from the
        credential certificate stored in the first data buffer in the array (credcert). Verify the validity of the
        certificates using Apple’s App Attest root certificate.

        See also: https://www.apple.com/certificateauthority/private/
        """
        cert = chain.pop(0)
        context = ValidationContext(extra_trust_roots=[self.attestation.config.root_ca])
        validator = CertificateValidator(cert, chain, validation_context=context)

        try:
            return validator.validate_usage({'digital_signature'})
        except (PathBuildingError, PathValidationError) as exception:
            raise InvalidCertificateChainException from exception

    def _get_extension(self, name: str, cert: Certificate) -> Extension:
        """ Helper method to get a specific x509 extension from a given certificate. """
        for extension in cert['tbs_certificate']['extensions']:
            if extension['extn_id'].native == name:
                return extension

        raise ExtensionNotFoundException



from typing import Optional, List

from pyattest.configs.config import Config
from pyattest.verifiers.google import GoogleVerifier


class GoogleConfig(Config):
    verifier_class = GoogleVerifier

    def __init__(self, key_ids: List[bytes], apk_package_name: str, production: bool = False,
                 root_cn: str = 'attest.android.com', root_ca: bytes = None):
        self.key_ids = key_ids
        self.apk_package_name = apk_package_name
        self.production = production
        self.root_cn = root_cn
        self._custom_root_ca = root_ca

    @property
    def root_ca(self) -> Optional[bytes]:
        """ This is only used for simplified unit testing. """
        return self._custom_root_ca



import datetime
import struct
from hashlib import sha256
from pathlib import Path

from _cbor2 import dumps as cbor_encode
from asn1crypto.core import OctetString
from cryptography import x509
from cryptography.hazmat._oid import ObjectIdentifier
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization.base import load_pem_private_key
from cryptography.x509.base import load_pem_x509_certificate
from cryptography.x509.extensions import UnrecognizedExtension
from cryptography.x509.oid import NameOID

from pyattest.testutils.factories.certificates import key_usage


def get(app_id: str, nonce: bytes, aaguid: bytes = b'appattestdevelop', counter: int = 0,
        wrong_public_key: bool = False):
    """ Helper to create a fake apple attestation. """
    root_key = load_pem_private_key(Path('pyattest/testutils/fixtures/root_key.pem').read_bytes(), b'123')
    root_cert = load_pem_x509_certificate(Path('pyattest/testutils/fixtures/root_cert.pem').read_bytes())

    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
    )

    public_key = private_key.public_key().public_bytes(encoding=serialization.Encoding.DER,
                                                       format=serialization.PublicFormat.PKCS1)

    auth_data_public_key = public_key if not wrong_public_key else 'XXXXX'.encode()

    auth_data = (sha256(app_id.encode()).digest()
                 + b'\x00'  # Flag, we'll fill it with zero
                 + struct.pack('!I', counter)
                 + aaguid  # TODO: Fillup to ensure 16 bytes
                 + struct.pack('!H', 32)  # Our SHA-256 digest is always 32bytes
                 + sha256(auth_data_public_key).digest())

    nonce = sha256(auth_data + sha256(nonce).digest())

    # When comparing this nonce with the one calculated on the server, we'll strip 6 bytes which are normally
    # used to indicated an ASN1 envelope sequence. See the verify_nonce method in the apple verifier.
    der_nonce = bytes(6) + OctetString(nonce.digest()).native

    subject = x509.Name([x509.NameAttribute(NameOID.ORGANIZATION_NAME, 'pyattest-testing-leaf')])
    cert = x509.CertificateBuilder() \
        .subject_name(subject) \
        .issuer_name(root_cert.subject) \
        .public_key(private_key.public_key()) \
        .serial_number(x509.random_serial_number()) \
        .not_valid_before(datetime.datetime.utcnow()) \
        .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=10)) \
        .add_extension(key_usage, critical=False) \
        .add_extension(UnrecognizedExtension(ObjectIdentifier('1.2.840.113635.100.8.2'), der_nonce), critical=False) \
        .sign(root_key, hashes.SHA256())

    data = {
        'fmt': 'apple-appattest',
        'attStmt': {
            'x5c': [
                cert.public_bytes(serialization.Encoding.DER),
                root_cert.public_bytes(serialization.Encoding.DER),
            ],
            'receipt': b'',
        },
        'authData': auth_data,
    }

    return cbor_encode(data), public_key



import base64
from typing import List, Optional, Tuple

import jwt
from asn1crypto import pem
from certvalidator import CertificateValidator, ValidationContext
from certvalidator.errors import PathValidationError, PathBuildingError
from certvalidator.path import ValidationPath
from jwt import InvalidTokenError

from pyattest.exceptions import PyAttestException, InvalidCertificateChainException, InvalidNonceException, \
    InvalidAppIdException, InvalidBasicIntegrity, InvalidCtsProfile, InvalidKeyIdException
from pyattest.verifiers.verifier import Verifier


class GoogleVerifier(Verifier):
    def verify(self):
        """
        Verify the given attestation based on the Google documentation. The attestation is formatted as JWS object.

        Raises a PyAttestException as soon as one of the verification steps fails.

        See also:
        Google Documentation https://developer.android.com/training/safetynet/attestation
        JWS RFC https://tools.ietf.org/html/draft-ietf-jose-json-web-signature-36
        """
        certificate_chain, data = self.unpack(self.attestation.raw)
        self.verify_nonce(data.get('nonce', None))
        self.verify_apk_package_name(data.get('apkPackageName', None))

        # We'll only validate basic integrity and cts profile on production
        if self.attestation.config.production:
            self.verify_basic_integrity(data.get('basicIntegrity', False))
            self.verify_cts_profile(data.get('ctsProfileMatch', False))

        self.verify_key_ids(data.get('apkCertificateDigestSha256', False))

        self.attestation.verified_data({'data': data, 'certs': certificate_chain})

    def unpack(self, jwt_object: str) -> Tuple[ValidationPath, dict]:
        """
        We first need to get the unverified headers to get a hold of the certificate so we can use the certs
        public key to verify the jwt objects signature. The certificate chain needs to be validated first.
        """
        header = jwt.get_unverified_header(jwt_object)
        certificate_chain = self._get_certificates(header)
        public_key = pem.armor('PUBLIC KEY', certificate_chain[-1].public_key.dump())

        try:
            data = jwt.decode(jwt_object, public_key, algorithms=['RS256'])
        except InvalidTokenError as exception:
            raise PyAttestException from exception

        return certificate_chain, data

    def verify_key_ids(self, key_ids: Optional[list]):
        """
        The apkCertificateDigestSha256 can hold multiple certificates and we're comparing them against the list
        of key ids in our config.
        """
        if not key_ids or len(key_ids) != len(self.attestation.config.key_ids):
            raise InvalidKeyIdException

        for key, config_key in zip(key_ids, self.attestation.config.key_ids):
            if key.encode() != config_key:
                raise InvalidKeyIdException

    def verify_cts_profile(self, cts_profile: bool):
        if not cts_profile:
            raise InvalidCtsProfile

    def verify_basic_integrity(self, basic_integrity: bool):
        if not basic_integrity:
            raise InvalidBasicIntegrity

    def verify_apk_package_name(self, package_name: Optional[str]):
        if not package_name or package_name != self.attestation.config.apk_package_name:
            raise InvalidAppIdException

    def verify_nonce(self, nonce: str):
        if base64.b64decode(nonce) != self.attestation.nonce:
            raise InvalidNonceException

    def _get_certificates(self, header: dict) -> ValidationPath:
        """
        Extract the SSL certificate chain from the JWS message. Before returning it, we need verify it otherwise
        it couldn't be used to check the jwt signature.
        """
        chain = list(map(base64.b64decode, header.get('x5c', [])))
        return self.verify_certificate_chain(chain)

    def verify_certificate_chain(self, chain: List[bytes]) -> ValidationPath:
        """
        Validate the SSL certificate chain and use SSL hostname matching to verify that the leaf certificate was
        issued to the hostname attest.android.com.

        We'll also allow to inject custom root ca for unittesting.
        """
        context = None
        if self.attestation.config.root_ca:
            context = ValidationContext(extra_trust_roots=[self.attestation.config.root_ca])

        validator = CertificateValidator(chain[0], chain[1:], validation_context=context)

        try:
            return validator.validate_tls(self.attestation.config.root_cn)
        except (PathBuildingError, PathValidationError) as exception:
            raise InvalidCertificateChainException from exception
